"""CLI command to download project results."""
import os
import re
import uuid
from collections import namedtuple

try:
    # python 3.7
    from urllib.parse import urlparse, parse_qs  # noqa
except ImportError:
    # python 2.7
    from urlparse import urlparse, parse_qs  # noqa

import requests  # noqa: I100

from tqdm import tqdm

from gencove import client  # noqa: I100
from gencove.constants import SAMPLE_STATUSES
from gencove.logger import echo, echo_debug, echo_warning
from gencove.utils import login


ALLOWED_STATUSES_RE = re.compile(
    "{}|{}".format(SAMPLE_STATUSES.succeeded, SAMPLE_STATUSES.failed),
    re.IGNORECASE,
)
FILENAME_RE = re.compile("filename=(.+)")
KILOBYTE = 1024
MEGABYTE = 1024 * KILOBYTE
NUM_MB_IN_CHUNK = 3
CHUNK_SIZE = NUM_MB_IN_CHUNK * MEGABYTE

Filters = namedtuple("Filters", ["project_id", "sample_ids", "file_types"])
Options = namedtuple("Options", ["host", "skip_existing"])


def download_deliverables(destination, filters, credentials, options):
    """Download project deliverables to a specified path on user machine.

    :param destination: path/to/save/deliverables/to.
    :type destination: str
    :param filters: allows to filter project deliverables to be downloaded
    :type filters: Filters
    :param host: API host to interact with.
    :type host: str
    :param credentials: login username/password
    :type credentials: Credentials
    :param options: different options to tweak execution
    :type options: Options
    """
    if not filters.project_id and not filters.sample_ids:
        echo_warning(
            "Must specify one of: project id or sample ids", err=True
        )
        return

    if filters.project_id and filters.sample_ids:
        echo_warning(
            "Must specify only one of: project id or sample ids", err=True
        )
        return

    echo_debug(
        "Host is {} downloading to {}".format(options.host, destination)
    )
    api_client = client.APIClient(options.host)
    login(api_client, credentials.email, credentials.password)

    if filters.project_id:
        echo_debug(
            "Retrieving sample ids for a project: {}".format(
                filters.project_id
            )
        )
        samples = api_client.get_project_samples(filters.project_id)[
            "results"
        ]
        echo_debug("Found {} project samples".format(len(samples)))

        if not samples:
            echo_warning("Project has no samples to download")
            return

        for sample in samples:
            _process_sample(
                destination,
                sample["id"],
                filters.file_types,
                api_client,
                options.skip_existing,
            )

        return

    for sample_id in filters.sample_ids:
        _process_sample(
            destination,
            sample_id,
            filters.file_types,
            api_client,
            options.skip_existing,
        )


def _download_file(download_to, file_prefix, url, skip_existing):
    """Download a file to file system.

    :param download_to: system/path/to/save/file/to
    :type download_to: str
    :param file_prefix: <client id>/<gencove sample id> to nest downloaded file
    under.
    :type file_prefix: str
    :param url: signed url from S3 to download the file from.
    :type url: str
    :param skip_existing: skip downloading existing files
    :type skip_existing: bool
    """
    with requests.get(url, stream=True) as req:
        req.raise_for_status()
        filename = _get_filename(req.headers["content-disposition"], url)
        filename_tmp = "download-{}.tmp".format(uuid.uuid4().hex)
        file_path = _create_filepath(download_to, file_prefix, filename)
        file_path_tmp = _create_filepath(
            download_to, file_prefix, filename_tmp
        )
        total = int(req.headers["content-length"])
        total_mb = int(total / MEGABYTE)

        # pylint: disable=C0330
        if (
            skip_existing
            and os.path.isfile(file_path)
            and os.path.getsize(file_path) == total
        ):
            echo("Skipping existing file: {}".format(file_path))
            return

        echo_debug("Starting to download file to: {}".format(file_path))

        with open(file_path_tmp, "wb") as downloaded_file:
            # pylint: disable=C0330
            for chunk in tqdm(
                req.iter_content(chunk_size=CHUNK_SIZE),
                total=total_mb / NUM_MB_IN_CHUNK,
                unit="MB",
                leave=True,
                desc="Progress: ",
                unit_scale=NUM_MB_IN_CHUNK,
            ):
                downloaded_file.write(chunk)

        os.replace(file_path_tmp, file_path)
        echo("Finished downloading a file: {}".format(file_path))


def _create_filepath(download_to, file_prefix, filename):
    """Build full file path and ensure that directory structure exists.

    :param download_to: top level directory path
    :type download_to: str
    :param file_prefix: subdirectories structure to create under download_to.
    :type file_prefix: str
    :param filename: name of the file inside download_to/file_prefix structure.
    :type filename: str
    """
    path = os.path.join(download_to, file_prefix)
    os.makedirs(path, exist_ok=True)
    file_path = os.path.join(path, filename)
    echo_debug("Deduced full file path is {}".format(file_path))
    return file_path


def _get_filename(content_disposition, url):
    """Deduce filename from content disposition or url.

    :param content_disposition: Request header Content-Disposition
    :type content_disposition: str
    :param url: URL string
    :type url: str
    """
    filename_match = re.findall(FILENAME_RE, content_disposition)
    if not filename_match:
        echo_debug(
            "Content disposition had no filename. Trying url query params"
        )
        filename = re.findall(FILENAME_RE, parse_qs(urlparse(url).query))
    else:
        filename = filename_match[0]
    if not filename:
        echo_debug(
            "URL didn't contain filename query argument. "
            "Assume filename from url"
        )
        filename = urlparse(url).path.split("/")[-1]
    echo_debug("Deduced filename to be: {}".format(filename))
    return filename


# pylint: disable=C0330
def _process_sample(
    destination, sample_id, file_types, api_client, skip_existing
):
    """Download sample deliverables."""
    sample = api_client.get_sample_details(sample_id)
    echo_debug(
        "Processing sample id {}, status {}".format(
            sample["id"], sample["last_status"]["status"]
        )
    )

    if not ALLOWED_STATUSES_RE.match(sample["last_status"]["status"]):
        echo_warning(
            "Sample #{} has no deliverable.".format(sample_id), err=True
        )
        return

    file_types_re = re.compile("|".join(file_types), re.IGNORECASE)

    for deliverable in sample["files"]:
        if file_types and not file_types_re.match(deliverable["file_type"]):
            echo_debug("Deliverable file type is not in desired file types")
            continue

        _download_file(
            destination,
            "{}/{}".format(sample["client_id"], sample["id"]),
            deliverable["download_url"],
            skip_existing=skip_existing,
        )



"""Tests for utils of Gencove CLI."""
from gencove.utils import upload_file


def test_upload_file(mocker):
    """Sanity check upload function."""
    mocked_s3_client = mocker.Mock()
    assert upload_file(
        mocked_s3_client,
        "foo.txt",
        "foo-bucket",
        object_name="foo-object.txt",
    )
    mocked_s3_client.upload_file.assert_called_once()



"""Utilities for different levels of click.echo outputs."""
import os

import click


INFO = "INFO"
DEBUG = "DEBUG"
LOG_LEVEL = os.environ.get("GENCOVE_LOG", INFO)


def output_warning(text):
    """Click echo warning."""
    return click.style(
        "WARNING: {}".format(text), bg="red", fg="white", bold=True
    )


def echo(msg, **kwargs):
    """Output click echo msg."""
    click.echo(msg, **kwargs)


def echo_debug(msg, **kwargs):
    """Output click echo msg only if debug is on."""
    if LOG_LEVEL == DEBUG:
        click.echo(msg, **kwargs)


def echo_warning(msg, **kwargs):
    """Output click echo msg with background."""
    echo(output_warning(msg), **kwargs)



"""Gencove CLI utils."""
import os

import boto3
from boto3.s3.transfer import TransferConfig

from botocore.credentials import RefreshableCredentials
from botocore.exceptions import ClientError
from botocore.session import get_session

import click

from gencove.constants import FASTQ_EXTENSIONS
from gencove.logger import echo_debug

KB = 1024
MB = KB * 1024
GB = MB * 1024


def get_s3_client_refreshable(refresh_method):
    """Return thread-safe s3 client with refreshable credentials.

    :param refresh_method: function that can get fresh credentials
    """
    session = get_session()
    session_credentials = RefreshableCredentials.create_from_metadata(
        metadata=refresh_method(),
        refresh_using=refresh_method,
        method="sts-assume-role",
    )
    # pylint: disable=protected-access
    session._credentials = session_credentials
    boto3_session = boto3.Session(botocore_session=session)
    return boto3_session.client(
        "s3", endpoint_url=os.environ.get("GENCOVE_LOCALSTACK_S3_ENDPOINT")
    )


def upload_file(s3_client, file_name, bucket, object_name=None):
    """Upload a file to an S3 bucket.

    :param s3_client: Boto s3 client
    :param file_name: File to upload
    :param bucket: Bucket to upload to
    :param object_name: S3 object name. If not specified then file_name is used
    :return: True if file was uploaded, else False
    """
    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_name

    # Upload the file
    try:
        # Set desired multipart threshold value of 5GB
        config = TransferConfig(
            multipart_threshold=100 * MB,
            multipart_chunksize=100 * MB,
            use_threads=True,
            max_concurrency=10,
        )
        s3_client.upload_file(file_name, bucket, object_name, Config=config)
    except ClientError as err:
        click.echo(
            "Failed to upload file {}: {}".format(file_name, err), err=True
        )
        return False
    return True


def seek_files_to_upload(path, path_root=""):
    """Generate a list of valid fastq files."""
    for root, dirs, files in os.walk(path):
        files.sort()
        for file in files:
            file_path = os.path.join(path_root, root, file)

            if file_path.lower().endswith(FASTQ_EXTENSIONS):
                echo_debug("Found file to upload: {}".format(file_path))
                yield file_path

        dirs.sort()
        for folder in dirs:
            seek_files_to_upload(folder, root)


def get_filename_from_path(path):
    """Cross OS get file name utility."""
    return os.path.normpath(path)


def login(api_client, email, password):
    """Login user into Gencove's system."""
    if not email or not password:
        click.echo("Login required")
        email = email or click.prompt("Email", type=str)
        password = password or click.prompt(
            "Password", type=str, hide_input=True
        )
    api_client.login(email, password)
    echo_debug("User logged in successfully")
    return api_client



"""Uploads fastq files to Gencove's system."""
import uuid
from datetime import datetime

from gencove import client
from gencove.constants import (
    FASTQ_EXTENSIONS,
    TMP_UPLOADS_WARNING,
    UPLOAD_PREFIX,
    UPLOAD_STATUSES,
)
from gencove.logger import echo, echo_debug, echo_warning
from gencove.utils import (
    get_filename_from_path,
    get_s3_client_refreshable,
    login,
    seek_files_to_upload,
    upload_file,
)


def upload_fastqs(source, destination, host, email, password):
    """Upload FASTQ files to Gencove's system.

    :param source: folder that contains fastq files to be uploaded.
    :type source: .fastq.gz, .fastq.bgz, .fq.gz, .fq.bgz
    :param destination: (optional) 'gncv://' notated folder
        on Gencove's system, where the files will be uploaded to.
    :type destination: str
    """
    echo_debug("Host is {}".format(host))

    files_to_upload = list(seek_files_to_upload(source))
    if not files_to_upload:
        echo(
            "No FASTQ files found in the path. "
            "Only following files are accepted: {}".format(FASTQ_EXTENSIONS),
            err=True,
        )
        return

    if destination and not destination.startswith(UPLOAD_PREFIX):
        echo(
            "Invalid destination path. Must start with '{}'".format(
                UPLOAD_PREFIX
            ),
            err=True,
        )
        return

    echo_warning(TMP_UPLOADS_WARNING, err=True)

    if not destination:
        destination = "{}cli-{}-{}".format(
            UPLOAD_PREFIX,
            datetime.utcnow().strftime("%Y%m%d%H%M%S"),
            uuid.uuid4().hex,
        )
        echo("Files will be uploaded to: {}".format(destination))

    api_client = client.APIClient(host)
    login(api_client, email, password)
    s3_client = get_s3_client_refreshable(api_client.get_upload_credentials)

    for file_path in files_to_upload:
        clean_filen_path = get_filename_from_path(file_path)
        gncv_notated_path = "{}/{}".format(destination, clean_filen_path)
        echo("Uploading {} to {}".format(file_path, gncv_notated_path))

        upload_details = api_client.get_upload_details(gncv_notated_path)
        if upload_details["last_status"]["status"] == UPLOAD_STATUSES.done:
            echo("File was already uploaded: {}".format(clean_filen_path))
            continue

        upload_file(
            s3_client=s3_client,
            file_name=file_path,
            bucket=upload_details["s3"]["bucket"],
            object_name=upload_details["s3"]["object_name"],
        )

    echo("All files were successfully synced.")



"""Describe all constants in Gencove CLI."""
from collections import namedtuple

HOST = "https://api.gencove.com"
TMP_UPLOADS_WARNING = (
    "The Gencove upload area is a temporary storage. "
    "Log into the dashboard and assign uploaded files "
    "to a project in order to avoid automatic deletion."
)


_APIEndpoint = namedtuple(
    "ApiEndpoint",
    [
        "get_jwt",
        "refresh_jwt",
        "verify_jwt",
        "upload_details",
        "get_upload_credentials",
        "project_samples",
        "sample_details",
    ],
)
API_ENDPOINTS = _APIEndpoint(
    "/api/v2/jwt-create/",
    "/api/v2/jwt-refresh/",
    "/api/v2/jwt-verify/",
    "/api/v2/uploads-post-data/",
    "/api/v2/upload-credentials/",
    "/api/v2/project-samples/{id}",
    "/api/v2/samples/{id}",
)

_UploadStatuses = namedtuple("UploadStatuses", ["done", "started", "failed"])
UPLOAD_STATUSES = _UploadStatuses("succeeded", "started", "failed")


FASTQ_EXTENSIONS = (".fastq.gz", ".fastq.bgz", ".fq.gz", ".fq.bgz")
UPLOAD_PREFIX = "gncv://"

_SampleStatuses = namedtuple("SampleStatuses", ["succeeded", "failed"])
SAMPLE_STATUSES = _SampleStatuses("succeeded", "failed")

Credentials = namedtuple("Credentials", ["email", "password"])



"""Python library which enables you to use Gencoves' research backend."""
import os

import click

from gencove import version
from gencove.commands.download import Filters, Options, download_deliverables
from gencove.commands.upload import upload_fastqs
from gencove.constants import Credentials, HOST
from gencove.logger import echo_debug


@click.group()
@click.version_option(version=version.version())
def cli():
    """Gencove's command line interface."""


@cli.command()
@click.argument("source")
@click.argument("destination", required=False)
@click.option(
    "--host",
    default=HOST,
    help="Optional Gencove API host, including http/s protocol. "
    "Defaults to https://api.gencove.com",
)
@click.option(
    "--email",
    default=lambda: os.environ.get("GENCOVE_EMAIL", ""),
    help="Gencove user email to be used in login. "
    "Can be passed as GENCOVE_EMAIL environment variable",
)
@click.option(
    "--password",
    default=lambda: os.environ.get("GENCOVE_PASSWORD", ""),
    help="Gencove user password to be used in login. "
    "Can be passed as GENCOVE_PASSWORD environment variable",
)
def upload(source, destination, host, email, password):
    """Upload FASTQ files to Gencove's system.

    :param source: folder that contains fastq files to be uploaded.
    :type source: .fastq.gz, .fastq.bgz, .fq.gz, .fq.bgz
    :param destination: (optional) 'gncv://' notated folder
        on Gencove's system, where the files will be uploaded to.
    :type destination: str

    Example:
        `gencove sync test_dataset gncv://test`
    """
    upload_fastqs(source, destination, host, email, password)


@cli.command()
@click.argument("destination")
@click.option("--project-id", help="Gencove project ID")
@click.option(
    "--sample-ids",
    help="A comma separated list of sample ids for "
    "which to download the deliverables",
)
@click.option(
    "--file-types",
    help="A comma separated list of deliverable file types to download.",
)
@click.option(
    "--host",
    default=HOST,
    help="Optional Gencove API host, including http/s protocol. "
    "Defaults to https://api.gencove.com",
)
@click.option(
    "--email",
    default=lambda: os.environ.get("GENCOVE_EMAIL", ""),
    help="Gencove user email to be used in login. "
    "Can be passed as GENCOVE_EMAIL environment variable",
)
@click.option(
    "--password",
    default=lambda: os.environ.get("GENCOVE_PASSWORD", ""),
    help="Gencove user password to be used in login. "
    "Can be passed as GENCOVE_PASSWORD environment variable",
)
@click.option(
    "--skip-existing/--no-skip-existing",
    default=True,
    help="Skip downloading files that already exist in DESTINATION",
)
def download(  # pylint: disable=C0330,R0913
    destination,
    project_id,
    sample_ids,
    file_types,
    host,
    email,
    password,
    skip_existing,
):  # noqa: D413 # pylint: disable=C0301
    """Download deliverables of a project.

    Must specify either project id or sample ids.

    :param destination: path/to/save/deliverables/to.
    :type destination: str
    :param project_id: project id in Gencove's system.
    :type project_id: str
    :param sample_ids: specific samples for which to download the results.
    if not specified, download deliverables for all samples.
    :type sample_ids: list(str)
    :param file_types: specific deliverables to download results for.
    if not specified, all file types will be downloaded.
    :type file_types: list(str)
    :param skip_existing: skip downloading existing files
    :type skip_existing: bool

    Examples:
        Download all samples results:

        `gencove download ./results --project-id d9eaa54b-aaac-4b85-92b0-0b564be6d7db`

        Download some samples:

        `gencove download ./results --sample-ids 59f5c1fd-cce0-4c4c-90e2-0b6c6c525d71,7edee497-12b5-4a1d-951f-34dc8dce1c1d`

        Download specific deliverables:

        `gencove download ./results --project-id d9eaa54b-aaac-4b85-92b0-0b564be6d7db --file-types alignment-bam,impute-vcf,fastq-r1,fastq-r2`
    """  # noqa: E501
    s_ids = tuple()
    if sample_ids:
        s_ids = tuple(s_id.strip() for s_id in sample_ids.split(","))
        echo_debug("Sample ids translation: {}".format(s_ids))

    f_types = tuple()
    if file_types:
        f_types = tuple(f_type.strip() for f_type in file_types.split(","))
        echo_debug("File types translation: {}".format(f_types))

    download_deliverables(
        destination,
        Filters(project_id, s_ids, f_types),
        Credentials(email, password),
        Options(host, skip_existing),
    )


if __name__ == "__main__":
    cli()



"""Packaging for Gencove CLI."""
from setuptools import find_packages, setup


def version():
    """Return current package version."""
    with open('gencove/version/A-major', 'rt') as f:
        major = f.read().replace('\n', '')
    with open('gencove/version/B-minor', 'rt') as f:
        minor = f.read().replace('\n', '')
    with open('gencove/version/C-patch', 'rt') as f:
        patch = f.read().replace('\n', '')

    return '{}.{}.{}'.format(major, minor, patch)


setup(
    name="gencove",
    description="Gencove API and CLI tool",
    url="http://docs.gencove.com",
    author="Tomaz Berisa",
    license="Apache 2.0",
    version=version(),
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        "Click>=7.0",
        "requests>=2.19.1",
        "boto3>=1.9.188",
        "future",
        "tqdm",
    ],
    setup_requires=[
        "pytest-runner"
    ],
    tests_require=[
        "pytest",
        "pytest-mock",
    ],
    entry_points="""
        [console_scripts]
        gencove=gencove.cli:cli
    """,
    package_data={"gencove": ["version/*"]},
)
